import { update } from "./update.js";

export const CellType = {
    SAND: 1,
    WALL: 2,
}

export type CellType = (typeof CellType)[keyof typeof CellType];

export type Cell = {
    type: CellType;
    moved: boolean;
    variation?: number;
} | undefined;

const WALL: Cell = {
    type: 2,
    moved: true,
};

export class StateBuffer {
    constructor(
        readonly buffer: Cell[],
        readonly width: number,
        readonly height: number
    ) {}

    get(x: number, y: number): Cell {
        if (x < 0 || y < 0 || x >= this.width || y >= this.height) {
            return WALL;
        }

        return this.buffer[y * this.width + x];
    }

    set(x: number, y: number, value: Cell) {
        if (x < 0 || y < 0 || x >= this.width || y >= this.height) return;

        this.buffer[y * this.width + x] = value;
    }

    fill(type: number) {
        if (type === 0) {
            this.buffer.fill(undefined);
        } else {
            throw new Error('fill for non-air is not yet implemented');
        }
    }
}

export class State {
    width: number;
    height: number;
    private bufferA: StateBuffer;
    private bufferB: StateBuffer;
    generation: number = 0;

    constructor(width: number, height: number);
    constructor(width: number, height: number, bufferA: StateBuffer, bufferB: StateBuffer, generation: number);
    constructor(
        width: number,
        height: number,
        bufferA?: StateBuffer,
        bufferB?: StateBuffer,
        generation?: number,
    ) {
        this.width = width;
        this.height = height;
        if (bufferA !== undefined && bufferB !== undefined) {
            this.bufferA = bufferA;
            this.bufferB = bufferB;
            this.generation = generation!;
        } else {
            this.bufferA = new StateBuffer(
                new Array(width * height).fill(0),
                width,
                height
            );
            this.bufferB = new StateBuffer(
                new Array(width * height).fill(0),
                width,
                height
            );
        }
    }

    get(): StateBuffer {
        return this.bufferA;
    }

    update(): State {
        update(this.bufferA, this.bufferB, this.generation);
        return new State(this.width, this.height, this.bufferB, this.bufferA, this.generation + 1);
    }
}
