import { Cell, CellType, StateBuffer } from "./state.js";


export function update(
    source: StateBuffer,
    target: StateBuffer,
    generation: number
) {
    const width = source.width;
    const height = source.height;

    target.fill(0);

    let swapped: [x: number, y: number][] = [];

    // First pass: fall trivial cells
    for (let y = 0; y < height; y++) {
        for (let x = 0; x < width; x++) {
            const cell = source.get(x, y);
            if (cell?.type === CellType.SAND) {
                if (isEmpty(source.get(x, y + 1))) {
                    cell.moved = true;
                    swapped.push([x, y]);
                    target.set(x, y + 1, cell);
                } else {
                    target.set(x, y, cell);
                }
            }
        }
    }

    // Second pass: try to move down cells using swapped
    while (swapped.length > 0) {
        const [x, y] = swapped.shift()!;
        if (!hasMoved(target.get(x, y - 1)) && isSand(target.get(x, y - 1))) {
            swap(target, x, y, x, y - 1);
            swapped.push([x, y - 1]);
        }
    }

    // Third pass: mark stuck cells (NOTE: this is the only step that is direction-dependent)
    function cannotMove(cell: Cell): boolean {
        if (!cell) return false;
        return cell.moved;
    }

    for (let y = height; y >= 0; y--) {
        for (let x = 0; x < width; x++) {
            const currentCell = target.get(x, y);
            if (!currentCell) continue;

            if (
                !currentCell.moved
                && cannotMove(target.get(x, y + 1))
                && cannotMove(target.get(x + 1, y + 1))
                && cannotMove(target.get(x - 1, y + 1))
            ) {
                currentCell.moved = true;
            }
        }
    }

    // Fourth pass: try to move cells diagonally
    for (let y = 0; y < height; y++) {
        for (let x = 0; x < width; x++) {
            const currentCell = target.get(x, y);
            if (!currentCell || currentCell.type !== CellType.SAND || hasMoved(currentCell)) {
                continue;
            }

            if (
                hasMoved(target.get(x, y + 1)) // || hasMoved(target.get(x, y + 2))
            ) {
                const left = isEmpty(target.get(x - 1, y + 1));
                const right = isEmpty(target.get(x + 1, y + 1));
                if (left && right && Math.random() < 0.5 || left && !right) {
                    swap(target, x, y, x - 1, y + 1);
                    swapped.push([x, y]);
                } else if (right) {
                    swap(target, x, y, x + 1, y + 1);
                    swapped.push([x, y]);
                }
            }
        }
    }

    // Fifth pass: try to move remaining cells to free spots
    while (swapped.length > 0) {
        const [x, y] = swapped.shift()!;
        if (isSand(target.get(x, y - 1))) {
            swap(target, x, y, x, y - 1);
            swapped.push([x, y - 1]);
        } else if (isSand(target.get(x - 1, y - 1)) && !hasMoved(target.get(x - 1, y - 1))) {
            swap(target, x, y, x - 1, y - 1);
            swapped.push([x - 1, y - 1]);
        } else if (isSand(target.get(x + 1, y - 1)) && !hasMoved(target.get(x + 1, y - 1))) {
            swap(target, x, y, x + 1, y - 1);
            swapped.push([x + 1, y - 1]);
        }
    }

    // Sixth pass: convert 2 -> 1
    for (let y = 0; y < height; y++) {
        for (let x = 0; x < width; x++) {
            const cell = target.get(x, y);
            if (cell?.moved) {
                cell.moved = false;
            }
        }
    }

    // Second pass: look for cells that could slot into swapped cells
    // do {
    //     while (swapped.length > 0) {
    //         const [x, y] = swapped.shift()!;
    //         if (target.get(x, y - 1) === 1) {
    //             target.set(x, y, 2);
    //             target.set(x, y - 1, 0);
    //             swapped.push([x, y - 1]);
    //         } else if (target.get(x - 1, y - 1) === 1) {
    //             if (target.get(x - 1, y) === 0) {
    //                 target.set(x - 1, y, 2);
    //                 target.set(x - 1, y - 1, 0);
    //                 swapped = swapped.filter(([x2, y2]) => x2 !== x - 1 || y2 !== y);
    //                 swapped.push([x - 1, y - 1]);
    //             } else if (target.get(x + 2, y) === 0 && Math.random() < 0.5) {
    //                 target.set(x + 2, y, 2);
    //                 target.set(x - 1, y - 1, 0);
    //                 swapped = swapped.filter(([x2, y2]) => x2 !== x - 2 || y2 !== y);
    //                 swapped.push([x, y]);
    //                 swapped.push([x - 1, y - 1]);
    //             } else {
    //                 target.set(x, y, 2);
    //                 target.set(x - 1, y - 1, 0);
    //                 swapped.push([x - 1, y - 1]);
    //             }
    //         } else if (target.get(x + 1, y - 1) === 1) {
    //             if (target.get(x + 1, y) === 0) {
    //                 target.set(x + 1, y, 2);
    //                 target.set(x + 1, y - 1, 0);
    //                 swapped = swapped.filter(([x2, y2]) => x2 !== x + 1 || y2 !== y);
    //                 swapped.push([x + 1, y - 1]);
    //             } else if (target.get(x - 2, y) === 0 && Math.random() < 0.5) {
    //                 target.set(x - 2, y, 2);
    //                 target.set(x + 1, y - 1, 0);
    //                 swapped = swapped.filter(([x2, y2]) => x2 !== x - 2 || y2 !== y);
    //                 swapped.push([x, y]);
    //                 swapped.push([x + 1, y - 1]);
    //             } else {
    //                 target.set(x, y, 2);
    //                 target.set(x + 1, y - 1, 0);
    //                 swapped.push([x + 1, y - 1]);
    //             }
    //         }
    //     }

    //     // Third pass: apply the diagonal movement rule to any unmoved cell
    //     for (let y = 0; y < height; y++) {
    //         for (let x = 0; x < width; x++) {
    //             if (target.get(x, y) === 1 && target.get(x, y + 1) > 0) {
    //                 const left = target.get(x - 1, y + 1) === 0;
    //                 const right = target.get(x + 1, y + 1) === 0;
    //                 if (left && right && Math.random() < 0.5 || left && !right) {
    //                     target.set(x - 1, y + 1, 2);
    //                     target.set(x, y, 0);
    //                     swapped.push([x, y]);
    //                 } else if (right) {
    //                     target.set(x + 1, y + 1, 2);
    //                     target.set(x, y, 0);
    //                     swapped.push([x, y]);
    //                 }
    //             }
    //         }
    //     }
    // } while (swapped.length > 0);
}



function isSand(cell: Cell): boolean {
    return cell?.type === 1;
}

function hasMoved(cell: Cell): boolean {
    return !!cell?.moved;
}

function isEmpty(cell: Cell): boolean {
    return cell === undefined;
}

function swap(buffer: StateBuffer, x1: number, y1: number, x2: number, y2: number) {
    function setMoved(cell: Cell): Cell {
        if (!cell) return cell;
        cell.moved = true;
        return cell;
    }

    const first = buffer.get(x1, y1);
    const second = buffer.get(x2, y2);

    buffer.set(x2, y2, setMoved(first));
    buffer.set(x1, y1, setMoved(second));
}
